/* gc-incremental.c
 * The incremental garbage collector.
 *
 * Copyright (c) 1997 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)
#define	FDBG(s)
#define	ADBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "gc.h"
#include "flags.h"
#include "classMethod.h"
#include "locks.h"
#include "thread.h"
#include "exception.h"
#include "external.h"
#include "lookup.h"
#include "md.h"

static gcList white;
static gcList grey;
static gcList black;
static gcList root;
static gcList threadstack;
static gcList finalise;

#if defined(GC_INCREMENTAL)
int32 gcAmount;
#else
static uint32 allocCount;
#endif


static struct {
	jword			free;
	jword			nextWidth;
	jword*			width[REF_MAXH];
} refTable;


/* We run the GC after allocating 1Mbyte of memory.  If we are
 * doing this incrementally, then we will have GC our entire heap
 * by the time we've allocated this much new space.
 */
#define	ALLOCCOUNTGC	(1024*1024)

static struct {
	iMux	mux;
	iCv	cv;
} finalman;
static struct {
	iMux	mux;
	iCv	cv;
} gcman;

int gc_mode = GC_DISABLED;	/* GC will be enabled after the first
				 * thread is setup.
				 */

static void* minMem = (void*)-1;
static void* maxMem = (void*)0;
static void* maxMemBin = (void*)0;

static struct {
	uint32	totalmem;
	uint32	totalobj;
	uint32	freedmem;
	uint32	freedobj;
	uint32	markedobj;
	uint32	markedmem;
	uint32	allocobj;
	uint32	allocmem;
} gcStats;

static void startGC(void);
static void finishGC(void);

/*
 * Initalise the Garbage Collection system.
 */
void
initGc(void)
{
	RESETLIST(white);
	RESETLIST(grey);
	RESETLIST(black);
	RESETLIST(root);
	RESETLIST(finalise);
	RESETLIST(threadstack);

DBG(	printf("root 0x%x\n", &root);
	printf("white 0x%x\n", &white);
	printf("grey 0x%x\n", &grey);
	printf("black 0x%x\n", &black);
	printf("final 0x%x\n", &finalise);
	printf("stack 0x%x\n", &threadstack);			)
}

/*
 * A reference from one object to another is being created.  Handle
 * the write barrier necessary for incremental collection.
 */
void
soft_addreference(void* f, void* t)
{
	gcInfo* from;
	gcInfo* to;

	/* Quit now for null objects */
	if (t == 0) {
		return;
	}
	from = MEM2GC(f);
	to = MEM2GC(t);

	if (from->colour == GC_BLACK && to->colour == GC_WHITE) {
		LOCK();
		to->colour = GC_GREY;
		REMOVELIST(to);
		APPENDLIST(grey, to);
		UNLOCK();
	}
}

/*
 * Walk a GC'able object.  Mark the object as black then for each pointer
 * in the object which refers to another GC object, grey the object
 * and insert into the grey list.
 */
void
walkObject(gcInfo* base)
{
	void** mem;
	gcInfo* obj;
	int32 i;

	assert(base->colour == GC_GREY);
	base->colour = GC_BLACK;
	if (base->final != GC_ROOT) {
		REMOVELIST(base);
		APPENDLIST(black, base);
	}

	gcStats.markedobj += 1;
	gcStats.markedmem += base->size;

DBG(	printf("walkObject: %d %x-%x\n", base->ref, GC2MEM(base),
		GC2MEM(base)+base->size);fflush(stdout);		)

	mem = GC2MEM(base);
	for (i = (base->size / sizeof(void*)) - 1; i >= 0; i--) {
		obj = MEM2GC(mem[i]);

		/* Check we've got a valid object */
		if (((void*)obj < minMem || (void*)obj > maxMem) ||
		    (((uintp)obj % sizeof(void*)) != 0) ||
		    (!VALIDREF(obj->ref)) ||
		    (GETREF(obj->ref) != (uintp)obj)) {
			continue;
		}
		/* If object's been traced before, don't do it again */
		if (obj->colour != GC_WHITE) {
			continue;
		}
		assert(obj->final != GC_ROOT);

		/* If we found a new white object, mark it as grey and
		 * move it into the grey list.
		 */
		LOCK();
		obj->colour = GC_GREY;
		REMOVELIST(obj);
		APPENDLIST(grey, obj);
		UNLOCK();
	}
}

/*
 * The Garbage Collector sits in a loop starting a collection, waiting
 * until it's finished incrementally, then tidying up before starting
 * another one.
 */
void
gcMan(void)
{
	gcInfo* obj;

	/* All threads start with interrupts disabled */
	intsRestore();

	_lockMutex(&gcman.mux);

	for(;;) {
#if defined(GC_INCREMENTAL)
		startGC();
		_waitCond(&gcman.mux, &gcman.cv, 0);

		LOCK();

		/* Walk the threadstack objects.  Although we do this
		 * initially it is possible that there's an object on the
		 * thread stack which isn't referene anywhere else but is
		 * live.
		 */
		for (obj = threadstack.next; obj != &threadstack; obj = obj->next) {
			obj->colour = GC_GREY;
			walkObject(obj);
			/* Keep the stats right */
			gcStats.markedobj -= 1;
			gcStats.markedmem -= obj->size;
		}
		/* Walk any remaining grey objects - shouldn't be necessary
		 * unless GC has been activated explicitly.
		 */
		for (obj = grey.next; obj != &grey; obj = grey.next) {
			walkObject(obj);
		}

		finishGC();

		UNLOCK();
#else
		_waitCond(&gcman.mux, &gcman.cv, 0);

		LOCK();

		startGC();
		for (obj = grey.next; obj != &grey; obj = grey.next) {
			walkObject(obj);
		}
		finishGC();

		UNLOCK();

#endif

		if (flag_gc) {
			fprintf(stderr, "<GC: total %dK, alloc %dK, marked %dK, freeing %dK>\n", gcStats.totalmem/1024, gcStats.allocmem/1024, gcStats.markedmem/1024, (gcStats.totalmem - gcStats.markedmem)/1024);
		}
		gcStats.allocobj = 0;
		gcStats.allocmem = 0;
	}
}

/*
 * Start the GC process by scanning the root and thread stack objects.
 */
static
void
startGC(void)
{
	gcInfo* obj;

	gcStats.freedmem = 0;
	gcStats.freedobj = 0;
	gcStats.markedobj = 0;
	gcStats.markedmem = 0;

	/* Walk the root objects */
	for (obj = root.next; obj != &root; obj = obj->next) {
		obj->colour = GC_GREY;
		walkObject(obj);
	}

	/* Walk the threadstack objects */
	for (obj = threadstack.next; obj != &threadstack; obj = obj->next) {
		obj->colour = GC_GREY;
		walkObject(obj);
	}
}

/*
 * Finish off the GC process.  Any unreached (white) objects are moved
 * for finalising and the finaliser woken.
 * The reached (black) objects are moved onto the now empty white list
 * and recoloured white.
 */
static
void
finishGC(void)
{
	gcInfo* obj;

	/* There shouldn't be any grey objects at this point */
	assert(grey.next == &grey);

	/* Any white objects should now be finalised */
	if (white.next != &white) {
		do {
			obj = white.next;
			REMOVELIST(obj);
			APPENDLIST(finalise, obj);
		} while (white.next != &white);

		_lockMutex(&finalman.mux);
		_signalCond(&finalman.cv);
		_unlockMutex(&finalman.mux);
	}

	/* Now move the black objects back to the white queue for next time.
	 */
	while (black.next != &black) {
		obj = black.next;
		assert(obj->colour == GC_BLACK);
		obj->colour = GC_WHITE;
		REMOVELIST(obj);
		APPENDLIST(white, obj);
	}
}

/*
 * The finaliser sits in a loop waiting to finalise objects.  When a
 * new finalised list is available, it is woken by the GC and finalises
 * the objects in turn.  An object is only finalised once after which
 * it is deleted.
 */
void
finaliserMan(void)
{
	gcInfo* obj;
	void* final;
	int i;

	/* All threads start with interrupts disabled */
	intsRestore();

	_lockMutex(&finalman.mux);

	for (;;) {
		_waitCond(&finalman.mux, &finalman.cv, 0);
		while (finalise.next != &finalise) {
			LOCK();
			obj = finalise.next;
			REMOVELIST(obj);
			UNLOCK();
			if (obj->final != GC_FINAL) {
				assert(obj->final != GC_ROOT);
				final = obj->final;
				/* Objects are only finalised once */
				obj->final = GC_FINAL;
				obj->colour = GC_WHITE;
				LOCK();
				/* Re-attach to the white list - it may
				 * still be live after finalizing.
				 */
				APPENDLIST(white, obj);
				UNLOCK();
				CALL_KAFFE_FUNCTION(((Method*)final), GC2MEM(obj));
			}
			else {
				SETREF(obj->ref, refTable.free);
				refTable.free = obj->ref;
FDBG(				printf("freeObject %x ref %d\n", obj,
					obj->ref);
					fflush(stdout);			)
				gcStats.totalmem -= obj->size;
				gcStats.totalobj -= 1;
				gcStats.freedmem += obj->size;
				gcStats.freedobj += 1;
				/* This horrible checking makes sure we never
				 * free the top of memory so we cannot
				 * reduce the heap size (which currently
				 * breaks things).
				 */
				if (obj == maxMem) {
					if (obj != maxMemBin) {
						if (maxMemBin != 0) {
							checked_free(maxMemBin);
						}
						maxMemBin = obj;
					}
					continue;
				}
				checked_free(obj);
			}
		}
	}
}

/*
 * Explicity invoke the garbage collector.
 */
void
invokeGC(void)
{
	_lockMutex(&gcman.mux);
	_signalCond(&gcman.cv);
	_unlockMutex(&gcman.mux);
}

/*
 * Allocate a new object.  The object is attached to the white queue.
 * After allocation, if incremental collection is active we peform
 * a little garbage collection.  If we finish it, we wakeup the garbage
 * collector.
 */
void*
gc_malloc(size_t size, void* final)
{
	gcInfo* obj;
	void* mem;
	int i;

	obj = calloc(size + sizeof(gcInfo), sizeof(uint8));
	if (obj == 0) {
		throwOutOfMemory();  
	}
	obj->size = size;

	/* Set min/max bounaries */
	if ((void*)obj > maxMem) {
		maxMem = obj;
	}
	if ((void*)obj < minMem) {
		minMem = obj;
	}
	obj->base = (void*)(obj+1);

	gcStats.totalmem += size;
	gcStats.totalobj += 1;
	gcStats.allocmem += size;
	gcStats.allocobj += 1;

	obj->final = final;
	mem = GC2MEM(obj);

	/* Allocate reference table as necessary */
	if (refTable.free == 0) {
		refTable.nextWidth++;
		assert(refTable.nextWidth < REF_MAXH);
		refTable.width[refTable.nextWidth] = checked_calloc(REF_MAXW, sizeof(uintp));
		for (i = 0; i < REF_MAXW - 1; i++) {
			refTable.width[refTable.nextWidth][i] = MAKEREF(refTable.nextWidth, i+1);
		}
		refTable.width[refTable.nextWidth][i] = MAKEREF(0,0);
		refTable.free = MAKEREF(refTable.nextWidth, 0);
	}
	/* Attach to reference table */
	obj->ref = refTable.free;
	refTable.free = GETREF(refTable.free);
	SETREF(obj->ref, obj);

	/* Insert new object into white list */
	LOCK();
	if (obj->final == GC_ROOT) {
		obj->colour = GC_BLACK;
		APPENDLIST(root, obj);
	}
	else {
#if defined(GC_INCREMENTAL)
		obj->colour = GC_GREY;
		APPENDLIST(grey, obj);
#else
		obj->colour = GC_WHITE;
		APPENDLIST(white, obj);
#endif
	}
	UNLOCK();

#if defined(GC_INCREMENTAL)
	/* Perform a little GC as we go */
	if (grey.next != &grey) {
		gcAmount += (gcStats.totalmem * size) / ALLOCCOUNTGC;
		while (gcAmount > 0) {
			gcAmount -= grey.next->size;
			walkObject(grey.next);
			if (grey.next == &grey) {
				invokeGC();
				break;
			}
		}
	}
#else
	allocCount += size;
	if (allocCount >= ALLOCCOUNTGC && gc_mode == GC_ENABLED) {
		allocCount = 0;
		invokeGC();
	}
#endif
	return (mem);
}

/*
 * Add a root object.
 */
void
gc_addroot(void* mem)
{
	gcInfo* obj;

	obj = MEM2GC(mem);
	assert(obj->final == GC_NORMAL);
	LOCK();
	REMOVELIST(obj);
	obj->final = GC_ROOT;
	obj->colour = GC_BLACK;
	APPENDLIST(root, obj);
	UNLOCK();
}

/*
 * Remove a root object.
 */
void
gc_rmroot(void* mem)
{
	gcInfo* obj;

	obj = MEM2GC(mem);
	assert(obj->final == GC_ROOT);
	LOCK();
	obj->final = GC_NORMAL;
	obj->colour = GC_GREY;
	REMOVELIST(obj);
	APPENDLIST(grey, obj);
	UNLOCK();
}

/*
 * Add a thread stack object.
 */
void
gc_addthreadstack(void* mem)
{
	gcInfo* obj;

	obj = MEM2GC(mem);
	assert(obj->final == GC_NORMAL);
	LOCK();
	REMOVELIST(obj);
	obj->final = GC_ROOT;
	obj->colour = GC_BLACK;
	APPENDLIST(threadstack, obj);
	UNLOCK();
}

/*
 * Create a phony threadstack object which maps a memory range.
 *  We use this to establish the initial thread's stack.
 */
void
gc_addthreadstack_phony(void* from, void* to)
{
	void* mem;
	gcInfo* obj;

	mem = gc_malloc(0, GC_NORMAL);
	obj = MEM2GC(mem);
	obj->base = from;
	obj->size = to - from;

	LOCK();
	REMOVELIST(obj);
	obj->final = GC_ROOT;
	obj->colour = GC_BLACK;
	APPENDLIST(threadstack, obj);
	UNLOCK();

	gcStats.totalmem += obj->size;
	gcStats.allocmem += obj->size;
}


/*
 * Remove a threadstack object.
 */
void
gc_rmthreadstack(void* mem)
{
	gcInfo* obj;

	obj = MEM2GC(mem);
	assert(obj->final == GC_ROOT);
	LOCK();
	obj->final = GC_NORMAL;
	obj->colour = GC_GREY;
	REMOVELIST(obj);
	APPENDLIST(grey, obj);
	UNLOCK();
}

/*
 * Create a new Java object.
 * If the object is a Class then we make it a root (for the moment).
 * Otherwise, if the object has a finalize method we look it up and store
 * it with the object for it's later use.  Otherwise the object is normal.
 */
void*
newObject(int sz, classes* class)
{
	object* obj;
	void* type;

        sz += sizeof(object);

	if (class == ClassClass) {
		type = GC_ROOT;
	}
	else if (class->final == false) {
		type = GC_NORMAL;
	}
	else {
		type = (void*)findMethod(class, final_name, void_signature);
	}
	obj = gc_malloc(sz, type);

        /* Fill in object information */
        obj->dtable = (class != 0 ? class->dtable : 0);

ADBG(	printf("newObject %x class %s ref %d\n", MEM2GC(obj),
		(class ? class->name->data : "<none>"),
		MEM2GC(obj)->ref);
		fflush(stdout);						)

        return (obj);
}
