/*
 * support.c
 * Native language support.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include <stdarg.h>
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "baseClasses.h"
#include "classMethod.h"
#include "lookup.h"
#include "errors.h"
#include "exception.h"
#include "slots.h"
#include "external.h"
#include "machine.h"
#include "support.h"
#include "md.h"
#include "itypes.h"

#define	MAXEXCEPTIONLEN		200
#define	ERROR_SIGNATURE		"(Ljava/lang/String;)V"

/* Anchor point for user defined properties */
userProperty* userProperties;

#if defined(NO_SHARED_LIBRARIES)
/* Internal native functions */
static nativeFunction null_funcs[1];
nativeFunction* native_funcs = null_funcs;
#endif

extern int strLengthUtf8 (char* str, int len);


/*
 * Convert an Java string to a C string.
 */
char*
javaString2CString(stringClass* js, char* cs, int len)
{
	jchar* chrs;

	chrs = (jchar*) ((js->value)+1) + js->offset;
	if (len <= 0) {
		return (0);
	}
	len--;
	if (len > js->count) {
		len = js->count;
	}
	cs[len] = 0;
	while (--len >= 0) {
		*cs++ = (char)*chrs++;
	}
	return (cs);
}

/*
 * Convert a Java string into a malloced C string buffer.
 */
char*
makeCString(stringClass* js)
{
	char* str;

	str = checked_malloc(js->count + 1);
	if (str != 0) {
		javaString2CString(js, str, js->count + 1);
	}
	return (str);
}

/*
 * Convert a C string into a Java String.
 */
stringClass*
makeJavaString(char* cs, int len)
{
	stringClass* obj;
	jchar* chrs;

	obj = (stringClass*)alloc_object(StringClass);
	/* FIXME - should intern string literals */
	obj->count = len;
	obj->value = alloc_array (len, TYPE_Char);
	obj->offset = 0;
	chrs = (jchar*) ((obj->value)+1);

	while (--len >= 0) {
		*chrs++ = *(unsigned char*)cs++;
	}
	return (obj);
}

stringClass*
makeJavaStringFromUtf8 (unsigned char* ptr, int len)
{
	stringClass* obj;
	register unsigned char* limit;
	jchar *chrs;
	int ch;

	limit = ptr + len;
	obj = (stringClass*)alloc_object(StringClass);
	/* FIXME - should intern string literals */
	obj->count = strLengthUtf8 (ptr, len);
	obj->value = alloc_array (len, TYPE_Char);
	obj->offset = 0;
	chrs = (jchar*) ((obj->value)+1);

	for (; ptr < limit;) {
		ch = UTF8_GET (ptr, limit);
		*chrs++ = ch;
	}

	return (obj);
}

stringClass*
makeReplaceJavaStringFromUtf8 (unsigned char* ptr, int len,
			       int from_ch, int to_ch)
{
	stringClass* obj;
	register unsigned char* limit;
	jchar* chrs;
	int ch;

	limit = ptr + len;
	obj = (stringClass*)alloc_object(StringClass);
	/* FIXME - should intern string literals */
	obj->count = strLengthUtf8 (ptr, len);
	obj->value = alloc_array (len, TYPE_Char);
	obj->offset = 0;
	chrs = (jchar*) ((obj->value)+1);

	for (; ptr < limit;) {
		ch = UTF8_GET (ptr, limit);
		if (ch == from_ch) {
			ch = to_ch;
		}
		*chrs++ = ch;
	}

	return (obj);
}

stringClass*
Utf8Const2JavaString (Utf8Const* str)
{
	return makeJavaStringFromUtf8 (str->data, str->length);
}

/*
 * Convert a C string into a Java char array.
 */
object*
makeJavaCharArray(char* cs, int len)
{
	object* obj;
	int i;

	obj = alloc_array (len, TYPE_Char);
	if (cs != NULL) {
		jchar *ptr = (jchar*) (obj+1);
		for (i = 0;  i < len;  i++) {
			*ptr++ = *(unsigned char*)cs++;
		}
	}
	return (obj);
}

/*
 * Call a Java method from native code.
 */
jword
do_execute_java_method(void* ee, object* obj, char* method_name, char* signature, methods* mb, int isStaticCall, ...)
{
	char* sig;
	int args;
	va_list argptr;
	jword retval;

	if (mb == 0 || mb->ncode == 0) {
		mb = findMethod(OBJECT_CLASS(obj),
				makeUtf8Const (method_name, -1),
				makeUtf8Const (signature, -1));
		if (mb == 0) {
			throwException(NoSuchMethodError);
		}
		/* Method must be non-static to invoke it here */
		if ((mb->accflags & ACC_STATIC) != 0) {
			throwException(NoSuchMethodError);
		}
	}

	/* Calculate number of arguments */
	sig = signature;
	args = sizeofSig(&sig, false);

	/* Make the call */
	va_start(argptr, isStaticCall);
 	CALL_KAFFE_FUNCTION_VARARGS(mb, obj, args, argptr, retval);
	va_end(argptr);

	return (retval);
}

/*
 * Call a Java static method on a class from native code.
 */
jword
do_execute_java_class_method(char* cname, char* method_name, char* signature, ...)
{
	char* sig;
	int args;
	va_list argptr;
	classes* class;
	methods* mb;
	jword retval;
	char cnname[CLASSMAXSIG];	/* Unchecked buffer - FIXME! */

	/* Convert "." to "/" */
	classname2pathname(cname, cnname);

	class = lookupClass(cnname);
	assert(class != 0);
	mb = findMethod(class,
			makeUtf8Const (method_name, -1),
			makeUtf8Const (signature, -1));
	if (mb == 0) {
		throwException(NoSuchMethodError);
	}
	/* Method must be static to invoke it here */
	if ((mb->accflags & ACC_STATIC) == 0) {
		throwException(NoSuchMethodError);
	}

	/* Calculate number of arguments */
	sig = signature;
	args = sizeofSig(&sig, false);

	/* Make the call */
	va_start(argptr, signature);
 	CALL_KAFFE_FUNCTION_VARARGS(mb, 0, args, argptr, retval);
	va_end(argptr);

	return (retval);
}

/*
 * Allocate an object and execute the constructor.
 */
object*
execute_java_constructor(void* ee, char* cname, classes* cc, char* signature, ...)
{
	int args;
	object* obj;
	char* sig;
	va_list argptr;
	methods* mb;
	char buf[MAXEXCEPTIONLEN];
	jword retval;

	if (cc == 0) {
		/* Convert "." to "/" */
		classname2pathname(cname, buf);

		cc = lookupClass (buf);
		assert(cc != 0);
	}

	initializeClass (cc);

	mb = findMethod(cc,
			constructor_name,
			signature==VOID_SIGNATURE ? void_signature
			: makeUtf8Const (signature, -1));
	if (mb == 0) {
		throwException(NoSuchMethodError);
	}

	obj = alloc_object(cc);
	assert(obj != 0);

	/* Calculate number of arguments */
	sig = signature;
	args = sizeofSig(&sig, false);

	/* Make the call */
	va_start(argptr, signature);
 	CALL_KAFFE_FUNCTION_VARARGS(mb, obj, args, argptr, retval);
	va_end(argptr);

	return (obj);
}

/*
 * Signal an error by creating the object and throwing the exception.
 */
void
SignalError(void* ee, char* cname, char* str)
{
	object* obj;

	obj = execute_java_constructor(ee, cname, 0, ERROR_SIGNATURE, makeJavaString(str, strlen(str)));
	throwException(obj);
}

/*
 * Convert a class name to a path name.
 */
void
classname2pathname(char* from, char* to)
{
	int i;

	/* Convert any '.' in name to '/' */
	for (i = 0; from[i] != 0; i++) {
		if (from[i] == '.') {
			to[i] = '/';
		}
		else {
			to[i] = from[i];
		}
	}
	to[i] = 0;
}

/*
 * Return current time in milliseconds.
 */
jlong
currentTime(void)
{
	jlong tme;

#if defined(HAVE_GETTIMEOFDAY)
	struct timeval tm;
	gettimeofday(&tm, 0);
	tme = (((jlong)tm.tv_sec * (jlong)1000) + ((jlong)tm.tv_usec / (jlong)1000));
#else
	tme = 0;
#endif
	return (tme);
}

/*
 * Set a property to a value.
 */
void
setProperty(void* properties, char* key, char* value)
{
	do_execute_java_method(0, properties, "put",
		"(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;",
		0, false,
		makeJavaString(key, strlen(key)),
		makeJavaString(value, strlen(value)));
}

/*
 * Allocate a new object of the given class name.
 */
object*
AllocObject(char* classname)
{
	classes* class = lookupClass (classname);
	assert(class != 0);

	return (alloc_object(class));
}

/*
 * Allocate a new array of a given size and type.
 */
object*
AllocArray(int sz, int type)
{
	return (alloc_array(sz, type));
}

/*
 * Allocate a new array of the given size and class name.
 */
object*
AllocObjectArray(int sz, char* classname)
{
	char sig[CLASSMAXSIG];  /* FIXME - fixed-sized unchecked buffer! */
	/* Build signature for array type */
	strcpy(sig, "[");
	strcat(sig, classname);

	return (alloc_objectarray(sz, sig));
}

#if defined(NO_SHARED_LIBRARIES)
/*
 * Register an user function statically linked in the binary.
 */
void
addNativeMethod(char* name, void* func)
{
	static int funcs_nr = 0;
	static int funcs_max = 0;

	/* If we run out of space, reallocate */
	if (funcs_nr >= funcs_max) {
		funcs_max += NATIVE_FUNC_INCREMENT;
		if (native_funcs != null_funcs) {
			native_funcs = checked_realloc(native_funcs, funcs_max * sizeof(nativeFunction));
		}
		else {
			native_funcs = checked_malloc(NATIVE_FUNC_INCREMENT * sizeof(nativeFunction));
		}
	}
	native_funcs[funcs_nr].name = checked_malloc(strlen(name) + 1);
	strcpy(native_funcs[funcs_nr].name, name);
	native_funcs[funcs_nr].func = func;
	funcs_nr++;
	native_funcs[funcs_nr].name = 0;
	native_funcs[funcs_nr].func = 0;
}
#endif
